<?php
/**
 * Author: [ CopyLian ]
 * Date: [ 2017.02.23 ]
 * Email: [ copylian@aikehou.com ]
 * Site: [ http://www.copylian.com ]
 * Description [ 微信第三方平台处理类 ]
 */
namespace wechat;

class Thirdplatform {

    //前缀
    private $prefix = 'test_';

    //接口域名
    private $url = 'https://api.weixin.qq.com/cgi-bin/';

    //授权地址
    private $auth_url = 'https://mp.weixin.qq.com/cgi-bin/';

    //获取code地址
    private $requestCodeURL = 'https://open.weixin.qq.com/connect/oauth2/authorize';

    //权限api的地址
    private $oauthApiURL = 'https://api.weixin.qq.com/sns';

    //APPID
    private $appid;

    //APPSECRET
    private $appsecret;

    //消息校验Token
    private $token;

    //消息加解密Key
    private $key;

    /**
     * [__construct 构造函数]
     */
    public function __construct($prefix = '') {

        //前缀
        $this->prefix = !empty($prefix) ? $prefix : $this->prefix;

        //初始化数据
        $this->appid     = config('thirdplatform.' . $this->prefix . 'thirdplatform_account.appid');
        $this->appsecret = config('thirdplatform.' . $this->prefix . 'thirdplatform_account.appsecret');
        $this->token     = config('thirdplatform.' . $this->prefix . 'thirdplatform_account.token');
        $this->key       = config('thirdplatform.' . $this->prefix . 'thirdplatform_account.key');
    }

    /**
    * [get_thirdplatform_auth 授权注册页面扫码授权：注意此URL必须放置在页面当中用户点击进行跳转，不能通过程序跳转，否则将出现“请确认授权入口页所在域名，与授权后回调页所在域名相同....”错误]
    * @param  [type]  $redirect_uri [扫码成功后的回调地址]
    * @param  integer $auth_type    [要授权的帐号类型，1-则商户扫码后，手机端仅展示公众号；2-表示仅展示小程序；3-表示公众号和小程序都展示。如果为未指定，则默认小程序和公众号都展示。第三方平台开发者可以使用本字段来控制授权的帐号类型。]
    * @return [type]                [description]
    */
    public function get_thirdplatform_auth($redirect_uri = '', $auth_type = 3){

        //请求地址
        $url = $this->auth_url . "componentloginpage?component_appid=" . $this->appid . "&pre_auth_code=" . $this->get_pre_auth_code() . "&redirect_uri=" . urlencode($redirect_uri) . "&auth_type=" . $auth_type;
        return $url;
    }

    /**
     * [get_query_auth 使用授权码获取授权信息]
     * @param  string $auth_code [授权码, 会在授权成功时返回给第三方平台，详见第三方平台授权流程说明]
     */
    public function get_query_auth($auth_code = ''){

        //请求地址
        $url = $this->url . "component/api_query_auth?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => $this->appid,
            "authorization_code"      => $auth_code
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return '';
        } else {
            return $res['authorization_info'];
        }
    }

    /**
     * [get_authorizer_info 获取授权方的帐号基本信息]
     * @param  string $auth_appid [授权方 appid]
     */
    public function get_authorizer_info($auth_appid = ''){

        //请求地址
        $url = $this->url . "component/api_get_authorizer_info?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => $this->appid,
            "authorizer_appid"        => $auth_appid
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return [];
        } else {
            return $res;
        }
    }

    /**
     * [get_authorizer_option 获取授权方选项信息]
     * @param  string $auth_appid [授权方 appid]
     * @param  string $option_name [选项名称：location_report(地理位置上报选项)、voice_recognize(语音识别开关选项)、customer_service(多客服开关选项)]
     */
    public function get_authorizer_option($auth_appid = '', $option_name = ''){

        //请求地址
        $url = $this->url . "component/api_get_authorizer_option?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => $this->appid,
            "authorizer_appid"        => $auth_appid,
            "option_name"             => $option_name
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return [];
        } else {
            return $res;
        }
    }

    /**
     * [set_authorizer_option 设置授权方选项信息]
     * @param  string $auth_appid [授权方 appid]
     * @param  string $option_name [选项名称：location_report(地理位置上报选项)、voice_recognize(语音识别开关选项)、customer_service(多客服开关选项)]
     * @param  string $option_value [选项值：location_report(0-无上报、1-进入会话时上报、2-每5s上报)、voice_recognize(0-关闭语音识别、1-开启语音识别)、customer_service(0-关闭多客服、1-开启多客服)]
     */
    public function set_authorizer_option($auth_appid = '', $option_name = '', $option_value = 0){

        //请求地址
        $url = $this->url . "component/api_set_authorizer_option?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => $this->appid,
            "authorizer_appid"        => $auth_appid,
            "option_name"             => $option_name,
            "option_value"            => $option_value
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return [];
        } else {
            return $res;
        }
    }

    /**
     * [clear_quota 第三方平台 API 调用次数清零]
     */
    public function clear_quota($component_appid = ''){

        //请求地址
        $url = $this->url . "component/clear_quota?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => !empty($component_appid) ? $component_appid : $this->appid
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return [];
        } else {
            return $res;
        }
    }

    /**
     * [get_agent_thirdplatform_auth description]
     * @param  string $appid    [授权第三方平台的appid]
     * @param  string $back_url [重定向地址URL]
     * @param  [type] $state    [重定向后会带上 state 参数，开发者可以填写任意参数值，最多 128 字节]
     * @param  string $scope    [授权作用域，拥有多个作用域用逗号（,）分隔]
     * @param  string $code     [授权code]
     */
    public function get_agent_thirdplatform_auth($appid = '', $back_url = '', $state = null, $scope = 'snsapi_userinfo', $code = ''){
        //如果不存在code，则去授权
        if(!isset($code) || empty($code)){
            //获取授权code
            $url = $this->getRequestCodeURL($appid, $back_url, $state, $scope);

            //返回数据
            $rdata['redirect_url'] = $url;
            $rdata['msg'] = '获取授权CODE成功';
            $rdata['code'] = 2;
            return $rdata;
        }

        //获取
        $access_token = $this->getAccessToken($appid, 'code', $code);
        if(isset($access_token['errcode'])) {
            //当前code已经被使用过了
            if($access_token['errcode'] == 40163){
                $rdata['code'] = 3;
                $rdata['back_url'] = $back_url;
                $rdata['msg'] = $access_token['errmsg'];
            } else {
                $rdata['code'] = 0;
                $rdata['msg'] = $access_token['errmsg'];
            }
            return $rdata;
        }

        //获取用户信息，如果是：snsapi_userinfo 则获取用户完整信息，如果是静默授权 snsapi_base 则只获取openid
        if($scope == 'snsapi_userinfo'){
            $userinfo = $this->getUserInfo($access_token);
            if(isset($userinfo['errcode'])) {
                $rdata['code'] = 0;
                $rdata['msg'] = $userinfo['errmsg'];
                return $rdata;
            }
        } else {
            $userinfo['openid'] = $access_token['openid'];
        }

        //设置用户信息值
        $userinfo['state'] = $state;
        $userinfo['code'] = $code;
        $userinfo['back_url'] = $back_url;
        $userinfo['unionid']  = (isset($userinfo['unionid']) && !empty($userinfo['unionid'])) ? $userinfo['unionid'] : '';

        //返回数据
        $rdata['code'] = 1;
        $rdata['data'] = $userinfo;
        $rdata['msg'] = '获取用户信息成功';
        return $rdata;
    }

    /**
     * [getRequestCodeURL 获取授权code的URL]
     * @param  string $appid        [公众号的 appid]
     * @param  [type] $redirect_uri [重定向地址，需要 urlencode，这里填写的应是服务开发方的回调地址]
     * @param  [type] $state        [重定向后会带上 state 参数，开发者可以填写任意参数值，最多 128 字节]
     * @param  string $scope        [授权作用域，拥有多个作用域用逗号（,）分隔]
     */
    public function getRequestCodeURL($appid = '', $redirect_uri, $state = null, $scope = 'snsapi_userinfo'){
        
        //参数
        $query = array(
            'appid'           => $appid,
            'redirect_uri'    => $redirect_uri,
            'response_type'   => 'code',
            'scope'           => $scope,
            'component_appid' => $this->appid
        );

        //处理state
        if(!is_null($state) && preg_match('/[a-zA-Z0-9]+/', $state)){
            $query['state'] = $state;
        }

        //组装url
        $query = http_build_query($query);

        //返回url
        return "{$this->requestCodeURL}?{$query}#wechat_redirect";
    }

    /**
     * [getAccessToken 获取access_token，用于后续接口访问]
     * @param  string $type [description]
     * @param  [type] $code [description]
     * @return [type]       [description]
     */
    public function getAccessToken($appid = '', $type = 'client', $code = null){
        
        //组装参数
        $query = array(
            'appid'                  => $appid,
            'component_appid'        => $this->appid,
            'component_access_token' => $this->get_component_access_token()
        );

        //判断类型
        switch ($type) {
            case 'client':
                $query['grant_type'] = 'client_credential';
                $url = "{$this->url}token";
                break;

            case 'code':
                $query['code'] = $code;
                $query['grant_type'] = 'authorization_code';
                $url = "{$this->oauthApiURL}/oauth2/component/access_token";
                break;
        }

        //组装url
        $query = http_build_query($query);
        $url = $url ."?". $query;

        //请求并返回数据
        $res = $this->https_get($url);
        return object_to_array(json_decode($res));
    }

    /**
     * 获取授权用户信息
     * @param  string $token acess_token
     * @param  string $lang  指定的语言
     * @return array         用户信息数据，具体参见微信文档
     */
    public function getUserInfo($token = [], $lang = 'zh_CN'){

        //组装参数
        $query = array(
            'access_token' => $token['access_token'],
            'openid'       => $token['openid'],
            'lang'         => $lang,
        );

        //组装url
        $url = "{$this->oauthApiURL}/userinfo";
        $query = http_build_query($query);
        $url = $url ."?". $query;

        //返回数据
        $res = $this->https_get($url);
        return object_to_array(json_decode($res));
    }

    /**
     * [get_authorizer_token 获取/刷新接口调用令牌]
     * @param  string $authorizer_appid         [第三方平台 appid]
     * @param  string $authorizer_refresh_token [授权方 appid]
     * @return [type]                           [刷新令牌，获取授权信息时得到]
     */
    public function get_authorizer_token($authorizer_appid = '', $authorizer_refresh_token = ''){

        //获取缓存
        $authorizer_access_token = cache('authorizer_access_token');

        //如果不存在缓存则新生成
        if(empty($authorizer_access_token)){

            //请求地址
            $url = $this->url . "component/api_authorizer_token?component_access_token=" . $this->get_component_access_token();

            //发送数据
            $data = [
                "component_appid"           => $this->appid,
                "authorizer_appid"          => $authorizer_appid,
                "authorizer_refresh_token"  => $authorizer_refresh_token
            ]; 

            //请求并返回数据
            $res = $this->https_post($url, json_encode($data));
            $res = object_to_array(json_decode($res));
            if(isset($res['errcode']) && $res['errcode'] != 0) {
                //报错，返回空
                $authorizer_access_token = [];
            } else {
                $authorizer_access_token = $res;
                cache('authorizer_access_token', $authorizer_access_token, 7000);
            }
        }

        //返回数据
        return $authorizer_access_token;
    }

    /**
     * [getSignPackage 获取签名信息]
     */
    public function getSignPackage($authorizer_appid = '', $ticket = '', $url = '') {
        //验证
        if(empty($authorizer_appid) || empty($ticket)){
            return [];
        }

        //票据
        $jsapiTicket = $ticket;

        // 注意 URL 一定要动态获取，不能 hardcode.
        if(empty($url)){
            $protocol = (!empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] !== 'off' || $_SERVER['SERVER_PORT'] == 443) ? "https://" : "http://";
            $url = "$protocol$_SERVER[HTTP_HOST]$_SERVER[REQUEST_URI]";
        } else {
            //被转义过的url与当前网址保持一致
            $url = htmlspecialchars_decode($url);
        }

        $timestamp = time();
        $nonceStr = $this->createNonceStr();

        // 这里参数的顺序要按照 key 值 ASCII 码升序排序
        $string = "jsapi_ticket=$jsapiTicket&noncestr=$nonceStr&timestamp=$timestamp&url=$url";
        $signature = sha1($string);
        $signPackage = array(
           "appId"     => $authorizer_appid,
           "nonceStr"  => $nonceStr,
           "timestamp" => $timestamp,
           "url"       => $url,
           "signature" => $signature,
           "rawString" => $string
        );
        return $signPackage; 
    }

    /**
     * [createNonceStr 创建随机数]
     * @param  integer $length [长度]
     */
    public function createNonceStr($length = 16) {
        //随机因子
        $chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

        //生成随机字符串
        $str = "";
        for ($i = 0; $i < $length; $i++) {
          $str .= substr($chars, mt_rand(0, strlen($chars) - 1), 1);
        }

        //返回数据
        return $str;
    }

    /**
     * [getJsApiTicket 获取微信JS接口的临时票据]
     * @param  string $authorizer_appid [授权authorizer_appid]
     */
    public function getJsApiTicket($authorizer_appid = '', $authorizer_refresh_token = '') {

        //获取/刷新接口调用令牌
        $access_token = $this->get_authorizer_token($authorizer_appid, $authorizer_refresh_token);
        if(empty($access_token)){
            $rdata['errcode'] = 52000;
            $rdata['errmsg'] = '获取access_token异常';
            return $rdata;
        }

        //组装参数
        $query = array(
            'access_token' => $access_token['authorizer_access_token'],
            'type'         => 'jsapi'
        );

        //组装url
        $url = "{$this->url}ticket/getticket";
        $query = http_build_query($query);
        $url = $url ."?". $query;

        //返回数据
        $res = $this->https_get($url);
        return object_to_array(json_decode($res));
    }

    /**
     * [decryptMsg 消息解密]
     */
    public function decryptMsg(){

        //时间戳
        $timeStamp = empty($_GET['timestamp']) ? '' : trim($_GET['timestamp']);

        //临时数据
        $nonce = empty($_GET['nonce']) ? '' : trim ($_GET['nonce']);

        //消息签名
        $msg_sign = empty($_GET['msg_signature']) ? "" : trim($_GET['msg_signature']);

        //获取加密的数据
        $encryptMsg = file_get_contents('php://input');

        //实例化加密类
        $pc = new \wechat\crypt\WXBizMsgCrypt($this->token, $this->key, $this->appid);

        //xml对象解析
        $postArr = xml2array($encryptMsg);

        //格式化xml
        $format = "<xml><ToUserName><![CDATA[toUser]]></ToUserName><Encrypt><![CDATA[%s]]></Encrypt></xml>";
        $from_xml = sprintf($format, $postArr['Encrypt']);

        //第三方收到公众号平台发送的消息
        $msg = '';

        //解密
        $errCode = $pc->decryptMsg($msg_sign, $timeStamp, $nonce, $from_xml, $msg);

        //返回数据
        if($errCode == 0) {
            //xml对象解析
            $data = xml2array($msg);
            return $data;
        } else {
            return [];
        }
    }

    /**
     * [response 发送消息]
     * @param  string $msg     [消息]
     * @param  array  $dataMsg [接受的原始数据]
     * @param  array  $dataMsg [消息类型]
     */
    public function response($msg = '', $dataMsg = [], $type = 'text'){

        //时间戳
        $timeStamp = time();

        //临时
        $nonce = time();

        /* 基础数据 */
        $data = array(
            'ToUserName'   => $dataMsg['FromUserName'],
            'FromUserName' => $dataMsg['ToUserName'],
            'CreateTime'   => $timeStamp,
            'MsgType'      => $type,
            'Content'      => $msg
        );

        //转换数据为XML
        $xml = new \SimpleXMLElement('<xml></xml>');
        self::data2xml($xml, $data);
        $xml = $xml->asXML();

        //加密
        $pc = new \wechat\crypt\WXBizMsgCrypt($this->token, $this->key, $this->appid);
        $encryptMsg = '';
        $pc->encryptMsg($xml, $timeStamp, $nonce, $encryptMsg);
        exit($encryptMsg);
    }

    /**
     * [custom_send 给指定用户推送信息]
     * @param  string $authorizer_access_token [access_token]
     * @param  array  $dataMsg                 [消息数据]
     * @param  string $type                    [类型]
     */
    public function custom_send($authorizer_access_token  ='', $dataMsg = [], $type = 'text'){
        
        //基础数据
        $send_data = array(
            'touser'  => $dataMsg['FromUserName'],
            'msgtype' => $type,
            'text'    => [
                'content' => $dataMsg['query_auth_code'] . '_from_api'
            ]
        );

        //file_put_contents('4.txt', json_encode($send_data), FILE_APPEND);

        //编码
        $send_data = urldecode(json_encode($send_data));

        //组装参数
        $query = array(
            'access_token' => $authorizer_access_token,
        );

        //组装url
        $url = "{$this->url}message/custom/send";
        $query = http_build_query($query);
        $url = $url ."?". $query;

        //file_put_contents('5.txt', json_encode($url), FILE_APPEND);

        //返回数据
        $res = $this->https_post($url, $send_data);
        return object_to_array(json_decode($res));
    }

    /**
     * [get_user_info 获取指定用户的详细信息]
     * @param  string $openid [openid]
     * @param  string $lang   [语言]
     */
    public function get_user_info($authorizer_appid = '', $authorizer_refresh_token = '', $openid = '', $lang = 'zh_CN'){

        //获取/刷新接口调用令牌
        $access_token = $this->get_authorizer_token($authorizer_appid, $authorizer_refresh_token);
        if(empty($access_token)){
            return [];
        }

        //组装参数
        $query = array(
            'access_token' => $access_token['authorizer_access_token'],
            'openid'       => $openid,
            'lang'         => $lang
        );

        //组装url
        $url = "{$this->url}user/info";
        $query = http_build_query($query);
        $url = $url ."?". $query;

        //返回数据
        $res = $this->https_get($url);
        return object_to_array(json_decode($res));
    }

    /**
     * 数据XML编码
     * @param  object $xml  XML对象
     * @param  mixed  $data 数据
     * @param  string $item 数字索引时的节点名称
     * @return string
     */
    protected static function data2xml($xml, $data, $item = 'item') {
        foreach ($data as $key => $value) {
            /* 指定默认的数字key */
            is_numeric($key) && $key = $item;

            /* 添加子元素 */
            if(is_array($value) || is_object($value)){
                $child = $xml->addChild($key);
                self::data2xml($child, $value, $item);
            } else {
                if(is_numeric($value)){
                    $child = $xml->addChild($key, $value);
                } else {
                    $child = $xml->addChild($key);
                    $node  = dom_import_simplexml($child);
                    $cdata = $node->ownerDocument->createCDATASection($value);
                    $node->appendChild($cdata);
                }
            }
        }
    }
    
    /**
     * [get_component_access_token 获取令牌：component_access_token，每个令牌是存在有效期（2小时）的，且令牌的调用不是无限制的，请第三方平台做好令牌的管理，在令牌快过期时（比如1小时50分）再进行刷新。所以要对component_access_token做好本地缓存]
     */
    private function get_component_access_token() {

        //保存 component_access_token
        $component_access_token = cache('component_access_token');

        //空的情况下生成缓存
        if(empty($component_access_token)){
            //令牌请求地址
            $url = $this->url . "component/api_component_token";

            //验证票据 component_verify_ticket
            $where = [];
            $where[] = ['id', '=', 1];
            $component_verify_ticket = \Db::name('wechat_thirdplatform_ticket')->where($where)->value('ticket');

            //发送数据
            $data = [
                "component_appid"         => $this->appid,
                "component_appsecret"     => $this->appsecret,
                "component_verify_ticket" => $component_verify_ticket
            ];

            //请求并返回数据
            $res = $this->https_post($url, json_encode($data));
            $res = object_to_array(json_decode($res));
            if(isset($res['errcode']) && $res['errcode'] != 0) {
                //报错，返回空
                $component_access_token = '';
            } else {
                //获取 component_access_token
                $component_access_token = $res['component_access_token'];

                //生成缓存
                cache('component_access_token', $component_access_token, 6600);
            }   
        }

        //返回数据
        return $component_access_token;
    }

    /**
    * [get_pre_auth_code 获取预授权码]
    */
    private function get_pre_auth_code() {

        //请求地址
        $url = $this->url . "component/api_create_preauthcode?component_access_token=" . $this->get_component_access_token();

        //发送数据
        $data = [
            "component_appid"         => $this->appid
        ]; 

        //请求并返回数据
        $res = $this->https_post($url, json_encode($data));
        $res = object_to_array(json_decode($res));
        if(isset($res['errcode']) && $res['errcode'] != 0) {
            //报错，返回空
            return '';
        } else {
            return $res['pre_auth_code'];
        }
    }

    /**
     * [https_post curl处理post]
     * @param  [type] $url  [地址]
     * @param  [type] $data [json数据]
     */
    private function https_post($url, $data) {

        //初始化curl
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        if (!empty($data)){
            curl_setopt($curl, CURLOPT_POST, 1);
            curl_setopt($curl, CURLOPT_POSTFIELDS, $data);
        }

        //设置选项
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        $output = curl_exec($curl);
        curl_close($curl);

        //返回数据
        return $output;
    }

    /**
     * [https_get curl处理get]
     * @param  [type] $url [地址]
     */
    private function https_get($url) {

        //初始化curl
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, TRUE);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, FALSE);
        curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, FALSE);
        curl_setopt($curl, CURLOPT_HEADER, FALSE) ;
        curl_setopt($curl, CURLOPT_TIMEOUT,60);
        if (curl_errno($curl)) {
            return 'Errno'.curl_error($curl);
        } else {
            $result = curl_exec($curl);
        }
        curl_close($curl);

        //返回数据
        return $result;
    }
}